/* Pieces of mars.c needed for for GCC */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "mem.h"
#include "root.h"
#include "mars.h"
#include "module.h"
#include "mtype.h"

Global global;

Global::Global()
{
    mars_ext = "d";
    sym_ext  = "d";

    obj_ext  = "o";

    copyright = "Copyright (c) 1999-2004 by Digital Mars";
    written = "written by Walter Bright";
    version = "v0.79";
    global.structalign = 8;

    memset(&params, 0, sizeof(Param));
}

char *Loc::toChars()
{
    OutBuffer buf;
    char *p;

    if (filename)
    {
	buf.printf("%s", filename);
    }

    if (linnum)
	buf.printf("(%d)", linnum);
    buf.writeByte(0);
    return (char *)buf.extractData();
}

Loc::Loc(Module *mod, unsigned linnum)
{
    this->linnum = linnum;
    this->filename = mod ? mod->srcfile->toChars() : NULL;
}

/**************************************
 * Print error message and exit.
 */

void error(Loc loc, const char *format, ...)
{
    char *p = loc.toChars();

    if (*p)
	printf("%s: ", p);
    mem.free(p);

    va_list ap;
    va_start(ap, format);
    printf("Error: ");
    vprintf(format, ap);
    va_end( ap );
    printf("\n");
    fflush(stdout);

    global.errors++;
}

/***************************************
 * Call this after printing out fatal error messages to clean up and exit
 * the compiler.
 */

void fatal()
{
#if 0
    *(char *)0 = 0;
#endif
    exit(EXIT_FAILURE);
}
